#include <stdio.h>
#include <string.h>
#include <ldap.h>

/* prototypes for command functions */
typedef void (CMD_FUNC)( LDAP *, int, char ** );
CMD_FUNC add, auth, delete, list, dread, first;
void init( LDAP **, int, char ** );
void next( LDAP * );
/* protypes for utility functons */
int parse_command( char *, int *, char ** );
char *strtok_quote( char *line, char *sep );
void help( void );
void print_entry( LDAP *ld, LDAPMessage *e );

#define MAXARGS	20

main( int argc, char **argv )
{
	char	line[BUFSIZ];	/* command line */
	int	cargc;		/* command argument count */
	char	*cargv[MAXARGS];/* command argument vector */
	LDAP	*ld;		/* LDAP session handle */
	int	cmd;

	ld = NULL;
	printf( "cmd: " );
	while ( fgets( line, sizeof(line), stdin ) != NULL ) {
		cmd = parse_command( line, &cargc, cargv );
		if ( ld == NULL && (cmd == 'a' || cmd == 'b' || cmd == 'd'
		    || cmd == 'f' || cmd == 'n' || cmd == 'u') ) {
			printf( "LDAP session must be initialized first\n" );
			help();
		} else {
			switch ( cmd ) {
			case 'a':	/* asynchronous add */
				add( ld, cargc, cargv );
				break;
			case 'b':	/* asynchronous bind */
				auth( ld, cargc, cargv );
				break;
			case 'd':	/* asynchronous delete */
				delete( ld, cargc, cargv );
				break;
			case 'f':	/* first search entry */
				first( ld, cargc, cargv );
				break;
			case 'i':	/* initialize */
				init( &ld, cargc, cargv );
				break;
			case 'l':	/* list children of an entry */
				list( ld, cargc, cargv );
				break;
			case 'n':	/* next search entry */
				next( ld );
				break;
			case 'r':	/* read an entry */
				dread( ld, cargc, cargv );
				break;
			case 'u':	/* unbind */
				ldap_unbind( ld );
				break;
			default:
				help();
				break;
			}
		}
		printf( "cmd: " );
	}
}

void
init( LDAP **ld, int cargc, char **cargv )
{
	if ( cargc != 3 ) {
		printf( "usage: i host port\n" );
		return;
	}
	if ( (*ld = ldap_init( cargv[1], atoi( cargv[2] ) )) == NULL ) {
		perror( "ldap_init" );
	}
}

void
add( LDAP *ld, int cargc, char **cargv ) 
{
	LDAPMod	**attrs;
	char	*dn;

	if ( cargc != 2 ) {
		printf( "usage: a ldiffile\n" );
		return;
	}
	if ( ldif2entry( cargv[1], &dn, &attrs ) != 0 ) {
		printf( "Could not convert ldif file %s to entry\n",
		   cargv[1] );
		return;
	}
	if ( ldap_add_s( ld, dn, attrs ) != LDAP_SUCCESS ) {
		ldap_perror( ld, "ldap_add_s" );
	}
	ldap_mods_free( attrs, 1 );
	free( dn );
}

int
ldif2entry( char *fname, char **dn, LDAPMod ***attrs )
{
	FILE	*fp;
	char	line[BUFSIZ];
	char	*type, *value;
	int	lineno;

	/* open the ldif file */
	if ( (fp = fopen( fname, "r" )) == NULL ) {
		printf( "Could not open ldif file %s\n", fname );
		return( -1 );
	}
	lineno = 0;
	/* first line should be "dn: distinguished name\n" */
	if ( fgets( line, sizeof(line), fp ) == NULL ) {
		printf( "LDIF file %s is empty\n", fname );
		fclose( fp );
		return( -1 );
	}
	lineno++;
	line[ strlen(line) - 1 ] = '\0';
	if ( ldif2typeval( line, &type, &value ) != 0 ) {
		printf( "Bad LDIF in %s: line %d\n", fname, lineno );
		return( -1 );
	}
	if ( (*dn = strdup( value )) == NULL ) {
		printf( "No more memory\n" );
		fclose( fp );
		return( -1 );
	}
	/* following lines are "type: value\n" */
	*attrs = NULL;
	while ( fgets( line, sizeof(line), fp ) != NULL ) {
		lineno++;
		line[ strlen(line) - 1 ] = '\0';
		/* parse the ldif line into type and value */
		if ( ldif2typeval( line, &type, &value ) != 0 ) {
			printf( "Bad LDIF line in %s: line %d\n", fname,
			    lineno );
			free( *dn );
			ldap_mods_free( *attrs, 1 );
			return( -1 );
		}
		/* merge the type and value into our attr list */
		if ( merge_typeval( attrs, type, value ) != 0 ) {
			printf( "Cannot add LDIF from file %s\n",
			   fname );
			free( *dn );
			ldap_mods_free( *attrs, 1 );
			return( -1 );
		}
	}
	return( 0 );
}

int
ldif2typeval( char *ldif, char **type, char **value )
{
	/* the line looks like "type: value" */
	if ( (*value = strchr( ldif, ':' )) == NULL ) {
		return( -1 );
	}
	*(*value)++ = '\0';
	while ( isspace( **value ) ) {
		(*value)++;
	}
	*type = ldif;
	return( 0 );
}

int
merge_typeval( LDAPMod ***attrs, char *type, char *value )
{
	int	i, j;

	/* find the attribute in the LDAPMod array to merge with */
	for ( i = 0; *attrs != NULL && (*attrs)[i] != NULL; i++ ){
		if ( strcasecmp( (*attrs)[i]->mod_type, type ) == 0 ) {
			break;
		}
	}

	/* not found - create slot for the first one */
	if ( *attrs == NULL ) {
		if ( (*attrs = (LDAPMod **) calloc( 1, 2 * sizeof(LDAPMod *)) )
		   == NULL ) {
			return( -1 );
		}
		i = 0;
	/* not found - create slot for the new one */
	} else if ( (*attrs)[i] == NULL ) {
		if ( (*attrs = (LDAPMod **) realloc( *attrs,
		   (i + 2) * sizeof(LDAPMod *) )) == NULL ) {
			return( -1 );
		}
	}

	/* we've got a new slot - create the new mod */
	if ( (*attrs)[i] == NULL ) {
		if ( ((*attrs)[i] = (LDAPMod *) malloc( sizeof(LDAPMod) ))
		   == NULL ) {
			return( -1 );
		}
		(*attrs)[i]->mod_op = 0;
		(*attrs)[i]->mod_type = strdup( type );
		if ( ((*attrs)[i]->mod_values = (char **) malloc(
		    2 * sizeof(char *) )) == NULL ) {
			return( -1 );
		}
		(*attrs)[i]->mod_values[0] = strdup( value );
		(*attrs)[i]->mod_values[1] = NULL;
		(*attrs)[i + 1] = NULL;
		return( 0 );
	}

	/* found an existing attribute - add a value to it */
	for ( j = 0; (*attrs)[i]->mod_values[j] != NULL; j++ ) {
		/* NULL */
	}
	(*attrs)[i]->mod_values = (char **) realloc( (*attrs)[i]->mod_values,
	   (j + 2) * sizeof(char *) );
	(*attrs)[i]->mod_values[j] = strdup( value );
	(*attrs)[i]->mod_values[j+1] = NULL;

	return( 0 );
}

void
auth( LDAP *ld, int cargc, char **cargv )
{
	if ( cargc != 3 ) {
		printf( "usage: b dn passwd\n" );
		return;
	}
	if ( ldap_simple_bind_s( ld, cargv[1], cargv[2] )
	   != LDAP_SUCCESS ) {
		ldap_perror( ld, "ldap_simple_bind_s" );
		return;
	}
}

void
delete( LDAP *ld, int cargc, char **cargv )
{
	if ( cargc != 2 ) {
		printf( "usage: d dn\n" );
		return;
	}
	if ( ldap_delete_s( ld, cargv[1] ) != LDAP_SUCCESS ) {
		ldap_perror( ld, "ldap_delete_s" );
		return;
	}
}

void
list( LDAP *ld, int cargc, char **cargv )
{
	LDAPMessage	*res, *e;
	char		*dn;
	char		*attrs[2];
	char		**rdns;

	if ( cargc != 2 ) {
		printf( "usage: l dn\n" );
		return;
	}
	attrs[0] = "c";
	attrs[1] = NULL;
	if ( ldap_search_s( ld, cargv[1], LDAP_SCOPE_ONELEVEL,
	   "(objectclass=*)", attrs, 0, &res ) != LDAP_SUCCESS ) {
		ldap_perror( ld, "ldap_search_s" );
		return;
	}
	for ( e = ldap_first_entry( ld, res ); e != NULL;
	   e = ldap_next_entry( ld, e ) ) {
		if ( (dn = ldap_get_dn( ld, e )) == NULL ) {
			printf( "ldap_get_dn failed\n" );
			continue;
		}
		if ( (rdns = ldap_explode_dn( dn, 0 )) != NULL ) {
			printf( "%s\n", rdns[0] );
			ldap_value_free( rdns );
		}
		ldap_memfree( dn );
	}
	ldap_msgfree( res );
}

void
dread( LDAP *ld, int cargc, char **cargv )
{
	LDAPMessage	*res, *e;

	if ( cargc != 2 ) {
		printf( "usage: r dn\n" );
		return;
	}
	if ( ldap_search_s( ld, cargv[1], LDAP_SCOPE_BASE,
	   "(objectclass=*)", NULL, 0, &res ) != LDAP_SUCCESS ) {
		ldap_perror( ld, "ldap_search_s" );
		return;
	}
	if ( (e = ldap_first_entry( ld, res )) == NULL ) {
		printf( "Could not retrieve entry %s\n", cargv[1] );
		return;
	}
	print_entry( ld, e );
	ldap_msgfree( res );
}

void
print_entry( LDAP *ld, LDAPMessage *e )
{
	BerElement	*ber;
	char		*dn, *a;
	char		**vals;
	int		i;

	if ( (dn = ldap_get_dn( ld, e )) != NULL ) {
		printf( "dn: %s\n", dn );
		ldap_memfree( dn );
	}
	for ( a = ldap_first_attribute( ld, e, &ber ); a != NULL;
	   a = ldap_next_attribute( ld, e, ber ) ) {
		if ( (vals = ldap_get_values( ld, e, a )) != NULL ) {
			for ( i = 0; vals[i] != NULL; i++ ) {
				printf( "%s: %s\n", a, vals[i] );
			}
			ldap_value_free( vals );
		}
		ldap_memfree( a );
	}
	if ( ber != NULL ) {
		ber_free( ber, 0 );
	}
}

static int		srchid;

void
first( LDAP *ld, int cargc, char **cargv )
{
	char	*attrs[20];	/* max of 20 attributes by name */
	int	i, scope;

	if ( cargc < 4 ) {
		printf( "usage: f dn scope filter [attr ...]\n" );
		return;
	}
	/* determine scope */
	if ( strcasecmp( cargv[2], "sub" ) == 0 ) {
		scope = LDAP_SCOPE_SUBTREE;
	} else if ( strcasecmp( cargv[2], "one" ) == 0 ) {
		scope = LDAP_SCOPE_ONELEVEL;
	} else {
		printf( "Invalid scope %s\n", cargv[2] );
		return;
	}
	/* collect attributes to retrieve */
	for ( i = 4; i < cargc && i < 24; i++ ) {
		attrs[ i - 4 ] = cargv[i];
	}
	attrs[ i - 4 ] = NULL;
	/* initiate the search */
	if ( (srchid = ldap_search( ld, cargv[1], scope, cargv[3],
	   attrs[0] == NULL ? NULL : attrs, 0 )) < 0 ) {
		ldap_perror( ld, "ldap_search_s" );
		return;
	}

	next( ld );
}

void
next( LDAP *ld )
{
	LDAPMessage	*result, *e;

	if ( srchid == -1 ) {
		return;
	}
	if ( ldap_result( ld, srchid, 0, NULL, &result ) <= 0 ) {
		ldap_perror( ld, "ldap_result" );
		return;
	}
	if ( (e = ldap_first_entry( ld, result )) != NULL ) {
		print_entry( ld, e );
	} else {
		srchid = -1;
	}
	ldap_msgfree( result );
}

void
help( void )
{
	printf( "Commands:\n" );
	printf( "\ta\tadd an entry\n" );
	printf( "\tb\tbind to the directory\n" );
	printf( "\td\tdelete an entry\n" );
	printf( "\tf\tfirst search entry\n" );
	printf( "\ti\tinitialize an LDAP session\n" );
	printf( "\tl\tlist children of entry\n" );
	printf( "\tn\tnext search entry\n" );
	printf( "\tr\tread entry\n" );
	printf( "\tu\tunbind\n" );
}

int
parse_command( char *line, int *argcp, char **argv )
{
	char *	token;

	line[strlen(line)-1] = '\0';
	*argcp = 0;
	for ( token = strtok_quote( line, " \t" ); token != NULL;
	    token = strtok_quote( NULL, " \t" ) ) {
		if ( *argcp == MAXARGS ) {
			printf( "Too many arguments (max %d)\n", MAXARGS );
			exit( 1 );
		}
		argv[(*argcp)++] = token;
	}
	argv[*argcp] = NULL;
	return( *argcp == 0 ? -1 : *argv[0] );
}

char *
strtok_quote( char *line, char *sep )
{
	int		inquote;
	char		*tmp, *d;
	static char	*next;

	if ( line != NULL ) {
		next = line;
	}
	while ( *next && strchr( sep, *next ) ) {
		next++;
	}

	if ( *next == '\0' ) {
		next = NULL;
		return( NULL );
	}

	d = tmp = next;
	for ( inquote = 0; *next; next++ ) {
		switch ( *next ) {
		case '"':
			if ( inquote ) {
				inquote = 0;
			} else {
				inquote = 1;
			}
			break;
		case '\\':
			*d++ = *++next;
			break;
		default:
			if ( ! inquote ) {
				if ( strchr( sep, *next ) != NULL ) {
					*d++ = '\0';
					next++;
					return( tmp );
				}
			}
			*d++ = *next;
			break;
		}
	}
	*d = '\0';

	return( tmp );
}
